#!/usr/bin/python
###############################################################################
""" This script converts NetCDF file format to ASCII (columns of fixed width).

Uses 
 - ncdump for converting NetCDF into CDL
 - CDO splitname for splitting the NetCDF file into files with 1 variable
 - CDO outputf for writing ASCII columns of fixed width
 - zip for compression 
The number of written columns is equal to the length of the last coordinate
variable in the definition of the desired variable in the CDL header.

Version 0.3, Torsten Rathmann, Deutsches Klimarechenzentrum

Version 0.2: NetCDF with more than one variable can be converted.
Version 0.3: Bounds (help variables) do not cause error messages any longer.
             Outputs complete CDL-File
"""
###############################################################################

def shorten(s, delimiter):
    """ Shortens string s by string delimiter and the part right of delimiter
    and strip whitespace """
    i = s.find(delimiter)
    if i >= 0: s = s[:i]
    return s.strip()


def splitdim(dimnames, dimensions, s):
    """ Appends names of dimension variables found in string s to list
    dimnames, append dimensions found in s to list dimensions """
    s = shorten(s, ';')
    if len(s) > 0:
        strings = s.split(',')
        # split comma separated list: dimname1 = dim1, dimname2 = dim2, ...
        for s in strings:
            if s.find('=') < 1:  # syntax error
                return [], []
            substrings = s.split('=')
            dimname = substrings[0].strip()
            dimension = substrings[1].strip()
            if dimname == '' or dimension == '' or dimname.find(' ') > 0 or \
                dimension.find(' ') > 0:  # syntax error
                return [], []
            dimnames.append(dimname)
            dimensions.append(dimension)
    return dimnames, dimensions


def lastdim(cdlfile, variable):
    """ Finds last dimension of the data variable: 

    Reads the header of the CDL file given as argument. Looks for the
    definition of variable variable. If variable is found, their last
    dimension (dimension name and number) given in the variable defination 
    will be returned. If variable was not found or a syntax error in the
    header was found, an error message is returned.

    Reads CDL file
    Returns 3 strings: dimension name, dimension number, error message
    """

    dim = '' # dimension number of last declared dimension of searched variable

    f = open(cdlfile, 'r')

    dimnames = []
    dimensions = []

# Find Dimensions Part in CDL Header

    for line in f:
        strippedline = shorten(line,'//')  # eliminate comment
        if strippedline[0:11] == 'dimensions:':
            dimnames, dimensions = splitdim(dimnames, dimensions, \
                strippedline[11:])
            break
    else:  # end of file reached
        f.close()
        return '', '', '\'dimensions:\' missing'

# Read Dimension Names and Dimensions

    for line in f:
        strippedline = shorten(line,'//')    # eliminate comment
        if strippedline == '': continue      # ignore empty or comment line
        if strippedline.find('=') < 0: break # end of dimensions part found
        dimnames, dimensions = splitdim(dimnames, dimensions, strippedline)
    if dimensions == []:
        f.close()
        return '', '', 'Cannot read dimensions.'

# Find Variables Part in CDL Header

    if strippedline != 'variables:':
        for line in f:
            strippedline = shorten(line,'//')  # eliminate comment
            if strippedline == 'variables:': break
        else:  # end of file reached
            f.close()
            return '', '', '\'variables:\' missing'

# Find Variable in Variables Part

    for line in f:
        strippedline = shorten(line,'//')  # eliminate comment
        if strippedline == '': continue    # ignore empty or comment line
        words = strippedline.split()
        if words[0] in ['float', 'real', 'double', 'char', 'byte', \
            'short', 'int']:
            if len(words) < 2 or words[1][0] == '(':
                return '', '', 'Variable name is missing in line\r\n' \
                    + '\'' + line + '\''
            vardef = ''  # variable definition string
            for word in words[1:]: 
                vardef = vardef + word
            i = vardef.find('(')
            if i < 0:
                if vardef == variable: break
            else:
                if vardef[:i] == variable: break
    else:  # end of file reached
        f.close()
        return '', '', 'Variable ' + variable + ' not defined'

    f.close()

# Find Last Dimension in Dimensions List of Variable Definition

    if i < 0:  # scalar variable found
        return '1', '', ''
    j = vardef.find(')')
    if j < i:
        return '', '', '\')\' missing or in wrong position in line\r\n' \
            + '\'' + line + '\''
    words = vardef[i+1:j].split(',')
    # split comma separated dimension names and store them in words
    i = len(words)
    if i == 0:  # scalar variable found
        return '1', '', ''
    dimname = words[i-1]
    # The dimension name we are looking for is the last in the list.
    for i in range(len(dimnames)):
        if dimname == dimnames[i]:
            dim = dimensions[i]
            break
    if dim == '':
        return '', '', 'Dimension ' + dimname + ' was not properly ' + \
            'defined in dimensions part but used in variables part ' + \
            'in line\r\n' + '\'' + line + '\''

# Try to Read Real Dimension from Comment If Dimension Is Declared as Unlimited

    if dim.lower() == 'unlimited':
        f = open(cdlfile, 'r')
        for line in f:      # look for a comment like '// (4 currently)'
            words = line.lower().split()
            strippedline = ''
            for word in words:
                strippedline = strippedline + word
            if strippedline.find(dimname.lower() + '=unlimited;//(') > -1:
                i = strippedline.find('(')
                j = strippedline.find('currently)')
                dim = strippedline[i+1:j]
                break
        f.close()

# Check If Dimension Is Integer and Return

    if dim.isdigit():
        return dimname, dim, ''
    else:
        return '', '', 'Dimension is not integer.'


def datafilehelp(variable, dimname, dim):
    """ Creates help information about the formatted ASCII data """
    h = variable + '.txt\r\n' + \
        '  Data file containing variable ' + variable + '\r\n' + \
        '  ASCII, ' + dim + ' columns of fixed length\r\n'
    if dim == '1':
        h = h + '\r\n'
    else:
        h = h + \
            '  Each row contains ' + dim + ' values of ' + variable \
            + ' for the ' + dim + ' different\r\n' + \
            '  values of coordinate variable ' + dimname + ' printed in ' \
            + 'the file\r\n' + \
            '  all.cdl using the same sequence. The other coordinate(s)' \
            + '\r\n' + \
            '  is/are varied from row to row according to their sequence(s)' \
            + '\r\n' + \
            '  in all.cdl\r\n\r\n'
    return h


def errorhelp(path, cdlfile, zipfile, errormessage):
    """ Creates file error.txt with error information and appends it to Zip
    archive """
    h = 'The following internal error has occurred while reading ' + \
        cdlfile + ':\r\n\r\n' + errormessage + '\r\n\r\n' + \
        'Format conversion ceased.\r\n'
    f = open(path + 'error.txt', 'w')
    f.write(h)
    f.close()
    os.system('zip -j -g ' + zipfile + ' ' + path + 'error.txt ' + cdlfile)
    return


################################## Main Part ##################################

import os, shutil, sys, tempfile

# Check Command Argument

h = 'Converts NetCDF file format to ASCII with fixed column widths.\n' + \
    'Compresses ASCII data files, header file and help files using zip.\n\n' +\
    'Usage:   python ascii.py <NetCDF-file>\n'
if len(sys.argv) != 2 or sys.argv[1][0] == '-':
    print h
    sys.exit()

ncfile = sys.argv[1]  # name of NetCDF file 
if ncfile[-3:] != '.nc':
    print h
    raise IOError('NetCDF-file must have extension .nc')
if not os.path.isfile(ncfile):
    raise IOError('File ' + ncfile + ' does not exist.')

# Create Temporary Directory

path = tempfile.mkdtemp() + '/'      
# path is the absolute pathname of the new directory

# Set Filenames

zipfile = ncfile[:-3] + '.ascii.zip'  
# name of zip file, inclusive path of the original NetCDF file
cdlfile = path + 'all.cdl'  # name of CDL file
helpfile = path + 'readme.txt'

# Split NetCDF File with Multiple Noncoordinate Variables

os.system('cdo splitname ' + ncfile + ' ' + path)
ncfiles = os.listdir(path)

# Convert from NetCDF to CDL

os.system('ncdump ' + ncfile + ' > ' + cdlfile)

# Delete NetCDF File Containing All Variables

os.remove(ncfile)

# Create Zip File

os.system('zip -j ' + zipfile + ' ' + cdlfile)

# Help Information

h = 'The following files have been created for you:\r\n\r\n' + \
    'all.cdl\r\n' + \
    '  CDL (Common Data for Language) is the ASCII equivalent of\r\n' + \
    '  the NetCDF file format (Network Common Data Form).\r\n' + \
    '  all.cdl contains the complete header information and\r\n' + \
    '  all data.\r\n\r\n'

# Convert NetCDF Variable Data to ASCII and Append to Zip File

for ncfile in ncfiles:
    variable = ncfile[:-3]             # variable name
    ncfile = path + ncfile
    cdlfile = path + variable + '.cdl' # name of CDL header file (one variable)
    txtfile = path + variable + '.txt' # name of data file (formatted ASCII)
    os.system('ncdump -h ' + ncfile + ' > ' + cdlfile)  # write CDL header
    dimname, dim, errormessage = lastdim(cdlfile, variable)
    if errormessage == '':
        os.remove(cdlfile)
        os.system('cdo outputf,%13.6g,' + dim + ' ' + ncfile + ' > ' + txtfile)
        # convert data from NetCDF to ASCII, C style %13.6g columns
        os.remove(ncfile)
        os.system('zip -j -g '+ zipfile + ' ' + txtfile)
        # append ASCII file to Zip archive
        os.remove(txtfile)
        h = h + datafilehelp(variable, dimname, dim)  # help info
    else:
        errorhelp(path, cdlfile, zipfile, errormessage)  # create error.txt
        shutil.rmtree(path)  # remove temporary directory inclusive content
        sys.exit()
    
# Complete Help Information and Append It to Zip File

h = h + \
    'readme.txt\r\n' + \
    '  This help file\r\n\r\n' + \
    'readme_office.doc\r\n' + \
    '  Hints for the import of *.txt data files into Microsoft Excel\r\n' + \
    '  and Open Office\r\n\r\n' + \
    'Please check the file all.cdl to ensure you have received the\r\n' + \
    'desired data. In this CDL file you can find a dimension part.\r\n' + \
    'Often users want a 2-dimensional data field for each variable.\r\n' + \
    'In this case, exactly two of the dimensions needed for one\r\n' + \
    'variable must consist of more than one value.\r\n\r\n' + \
    'If you have got too much data, you may repeat the search with\r\n' + \
    'tighter coordinate intervals. The information given in the\r\n' + \
    'data part of the CDL file should help you set appropriate\r\n' + \
    'limits for the coordinates.\r\n\r\n' + \
    'Missing values (fill values) are set to \'1e+20\' in the *.txt\r\n' + \
    'data files.\r\n'

f = open(helpfile, 'w')
f.write(h)
f.close()
os.system('zip -j -g '+ zipfile + ' ' + helpfile + \
    ' readme_office.pdf')

# Remove Temporary Directory Inclusive Content

shutil.rmtree(path)  
